unit uSudokuBase;

interface
uses windows, classes, sysutils;

const
{$IFDEF SUPER_SUDOKU}
  LGBLOC = 4;
  HTBLOC = 4;
  TOUS_LES_CANDIDATS = $FFFF; // Masque de bits dsignant les 9 valeurs possibles.
  MODE_AFFICHAGE_LETTRE = true;
{$ELSE}
  LGBLOC = 3;
  HTBLOC = 3;
  TOUS_LES_CANDIDATS = $1FF; // Masque de bits dsignant les 9 valeurs possibles.
  MODE_AFFICHAGE_LETTRE = false;
{$ENDIF}
  LARGEUR = LGBLOC*LGBLOC; // Largeur d'une grille de Sudoku
  HAUTEUR = HTBLOC*HTBLOC; // Hauteur d'une grille de Sudoku
  VALEUR_MIN = 1; // Plus petite valeur jouable
  VALEUR_MAX = LGBLOC*HTBLOC; // Plus grande valeur jouable.
  CELLULE_VIDE = 0; // Valeur d'une case vide.

type
  TSudokuClass = class of TSudokuBase;
  
  // Classe de base pour la gestion d'une grille de Sudoku.
  // L'ide ici est de fournir une classe abstraite implmentant les contraintes du Sudoku. Elle est
  // destine  tre drive afin de tester diffrentes mthodes de rsolution de grilles,
  // avec diffrentes implmentations.
  // En particulier, on pourra avoir une classe drive implmentant l'algorithme de base, cod pour un
  // maximum de lisibilit.
  // Puis une autre classe drive pourra fournir une autre implmentation du mme algorithme, mais qui
  // sera cette fois ci code pour tre la plus efficace possible. Donc probablement compltement
  // illisible !
  TSudokuBase = class(TPersistent)
  private
    FListeSolution : TStrings;

    // Vrifie si la zone de la grille <x1, y1>  <x2, y2> respecte les rgles du sudoku :
    // Elle ne doit pas contenir de valeurs plus d'une fois. Les cellules vides ne comptent pas.
    // Cette mthode sert  la dfinition initiale de la grille, ainsi qu'au joueur qui tente de rsoudre
    // la grille. Elle ne sert pas  la rsolution automatique de la grille. Elle n'a pas besoin d'tre
    // optimise pour ses performances.
    function VerifierZone(x1, y1, x2, y2 : integer) : boolean;

  protected
    // Dfinit la valeur de la cellule <X,Y>  <Valeur>.
    // Cette mthode ne vrifie pas si le coup est valide.
    procedure DefinirValeur(X, Y, Valeur : integer); virtual; abstract;

    // Dfinit la cellule <X,Y> comme tant une celule rvle.
    procedure DefinirRevelee(X, Y : integer); virtual; abstract;

    class procedure RegisterAlgorithme;
  public
    // Cre et initialise une nouvelle instance de la classe TSudokuBase.
    constructor Create; virtual;

    // Dtruit la classe et libre les ressources utilises.
    destructor Destroy; override;

    //**************************************************************************************************
    //*   Mthodes destines au dessin de la grille et au joueur humain.                               *
    //*   Elles ne servent pas  la rsolution de la grille. Elles n'ont pas besoins d'tre optimises *
    //*   au niveau de leur performances.                                                              *
    //**************************************************************************************************

    // Rinitialise la grille avant de commencer une nouvelle partie : Les cellules rvles sont
    // conserves. Les autres sont effaces.
    procedure Initialiser; virtual;

    // Dfinie la valeur de la cellule <x, y>  <valeur>. La mthode vrifie que le coup est valide.
    // Si ce n'est pas le cas, la cellule n'est pas modifie.
    procedure DefinirCellule(x, y : integer; valeur : integer); virtual;

    // Ajoute/supprime un marqueur de valeur possible pour la cellule <x, y>.
    // Cette mthode sert au joueur humain qui tente de rsoudre la grille.
    procedure MarkerCellule(x, y : integer; candidat : integer); virtual; abstract;

    // Indique si la cellule <X, Y> est une cellule rvle. Les cellules rvle ne peuvent pas tre
    // modifies et doivent tre dessine en noir.
    function IsRevelee(x, y : integer) : boolean; virtual; abstract;

    // Retourne la valeur de la cellule <X,Y>. Les cellules vides retournent CELLULE_VIDE.
    function ValeurCellule(X, Y : integer) : integer; virtual; abstract;

    // Indique si la valeur <Valeur> est dsigne comme tant un candidat pour la cellule <X,Y>.
    function IsCandidat(X, Y, Valeur : integer) : boolean; virtual; abstract;

    // Indique si la rsolution de la grille est termine, c'est a dire si toutes les cellules sont
    // remplies et qu'elles respectent bien les contraintes du Sudoku.
    function GrilleTerminee: boolean; virtual;

    // Vrifie si la grille est valide. La mthode retourne True si toutes les cellules dfinies
    // respectent bien les contraintes du Sudoku. Les cellules vides sont ignores.
    function VerifierGrille: boolean; virtual;

    // Transforme la grille actuelle en grille de dpart :
    // Les cellules remplies sont dfinies comme tant des cellules rvles.
    procedure JouerLaGrille; virtual;

    // Charge la grille depuis le fichier <filename>.
    procedure LoadFromFile(filename : string); virtual;

    // Enregistre la grille dans le fichier <filename>.
    procedure SaveToFile(Filename : string); virtual;

    // Retourne le nom de l'algorithme utilis pour faire la rsolution.
    class function GetAlgorithmName : string; virtual; abstract;

    // Efface le contenu de la grille pour qu'on puisse lui dfinir une nouvelle
    // grille.
    procedure EffacerGrille; virtual; abstract;

    //**************************************************************************************************
    //*   Mthodes destines  la rsolution automatique de la grille.                                 *
    //*   Elles doivent tre optimises pour avoir les performances optimales                          *
    //**************************************************************************************************

    // Initialise la grille avant de tanter sa rsolution.
    procedure InitialiserResolution; virtual;

    // Recherche l'ensemble des candidats possibles pour la grille complte.
    procedure RechercherTousLesCandidats; virtual; abstract;

    // Tente de rsoudre la grille depuis sa dfinition actuelle. En retour, la mthode renvoie :
    // - true : La grille a t rsolue avec succs !
    // - false : La grille ne possde pas de solution !
    function Resoudre(NbMaxSolution : integer) : boolean; virtual; abstract;

    // Contient la liste de toutes les solutions qui ont t trouves jusqu' prsent.
    property ListeSolution : TStrings read FListeSolution write FListeSolution;
  end;

function ListeAlgorithme : TStrings;

implementation
var
  FListeAlgorithme : TStringList;

function ListeAlgorithme : TStrings;
begin
  result := FListeAlgorithme;
end;
  
{ TSudokuBase }

// Cre et initialise une nouvelle instance de la classe TSudokuBase.
constructor TSudokuBase.Create;
begin
  FListeSolution := TStringList.Create;
end;

// Dfinie la valeur de la cellule <x, y>  <valeur>. La mthode vrifie que le coup est valide.
// Si ce n'est pas le cas, la cellule n'est pas modifie.
procedure TSudokuBase.DefinirCellule(x, y, valeur: integer);
var
  old : integer;
begin
  if (not IsRevelee(X, Y)) and // On ne modifie pas une cellule rvle.
    (Valeur<>ValeurCellule(x, y))
  then begin
    old := ValeurCellule(x, y); // On mmorise l'ancienne valeur, avant modification.
    DefinirValeur(x, y, valeur); // On joue le coup demand

    // A prsent, il faut vrifier si la grille est encore valide.
    if (not VerifierZone(0, y, LARGEUR-1, y)) or // On vrifie la ligne
       (not VerifierZone(x, 0, x, HAUTEUR-1)) or // la colonne
       (not VerifierZone((x div LGBLOC)*LGBLOC,
                         (y div HTBLOC)*HTBLOC,
                         (x div LGBLOC)*LGBLOC +LGBLOC-1,
                         (y div HTBLOC)*HTBLOC +HTBLOC-1)) // et le bloc.
    then DefinirValeur(x, y, old); // Si les 3 lments ne sont pas valide, on restaure l'ancienne valeur.
  end;
end;

// Dtruit la classe et libre les ressources utilises.
destructor TSudokuBase.Destroy;
begin
  FListeSolution.Free;
  inherited Destroy;
end;

// Indique si la rsolution de la grille est termine, c'est a dire si toutes les cellules sont
// remplies et qu'elles respectent bien les contraintes du Sudoku.
function TSudokuBase.GrilleTerminee: boolean;
var
  x, y : integer;
begin
  result := VerifierGrille; // La grille doit tre valide.

  for x := 0 to LARGEUR-1 do
  begin
    for y := 0 to HAUTEUR-1 do
    begin
      result := result and (ValeurCellule(x, y)<>CELLULE_VIDE);
    end;
  end;
end;

// Rinitialise la grille avant de commencer une nouvelle partie : Les cellules rvles sont
// conserves. Les autres sont effaces.
procedure TSudokuBase.Initialiser;
var
  x, y : integer;
begin
  for x := 0 to LARGEUR-1 do
  begin
    for y := 0 to HAUTEUR-1 do
    begin
      // On efface toutes les cellules qui ne sont pas rvles.
      if not IsRevelee(X, Y)
      then DefinirCellule(X, Y, CELLULE_VIDE);
    end
  end;
end;

procedure TSudokuBase.InitialiserResolution;
begin

end;

// Transforme la grille actuelle en grille de dpart :
// Les cellules remplies sont dfinies comme tant des cellules rvles.
procedure TSudokuBase.JouerLaGrille;
var
  x, y : integer;
begin
  for x := 0 to LARGEUR-1 do
  begin
    for y := 0 to HAUTEUR-1 do
    begin
      if ValeurCellule(x, y)<>CELLULE_VIDE
      then DefinirRevelee(X, Y);
    end;
  end;
  FListeSolution.Clear; // On efface galement la liste des solutions.
end;

procedure TSudokuBase.LoadFromFile(filename: string);
var
  lst : TStrings;
  x, y : integer;
  ligne : string;
  n : integer;
begin
  EffacerGrille;
  lst := TStringList.Create;
  try
    lst.LoadFromFile(filename);
    for y := 0 to HAUTEUR-1 do
    begin
      if y<lst.Count
      then ligne := lst[y]
      else ligne := '';

      for x := 0 to LARGEUR-1 do
      begin
        if x < length(ligne)
        then DefinirValeur(x, y, ord(ligne[x+1])-48)
        else DefinirValeur(x, y, CELLULE_VIDE);

        // On efface les marqueurs des candidats.
        for n := VALEUR_MIN to VALEUR_MAX do
        begin
          if IsCandidat(x, y, n)
          then MarkerCellule(x, y, n);
        end;
      end;
    end;
  finally
    lst.Free;
  end;
  JouerLaGrille;
end;

class procedure TSudokuBase.RegisterAlgorithme;
begin
  FListeAlgorithme.Add(ClassName);
  RegisterClass(self);
end;

procedure TSudokuBase.SaveToFile(Filename: string);
var
  x, y : integer;
  lst : TStrings;
  ligne : string;
begin
  lst := TStringList.Create;
  try
    for y := 0 to HAUTEUR-1 do
    begin
      SetLength(Ligne, LARGEUR);
      for x := 0 to LARGEUR-1 do
      begin
        Ligne[x+1] := chr(ValeurCellule(x, y) + 48);
      end;
      lst.Add(Ligne);
    end;
    lst.SaveToFile(filename);
  finally
    lst.Free;
  end;
end;

// Vrifie si la grille est valide. La mthode retourne True si toutes les cellules dfinies
// respectent bien les contraintes du Sudoku. Les cellules vides sont ignores.
function TSudokuBase.VerifierGrille: boolean;
var
  x, y : integer;
begin
  result := true;

  // On vrifie les colonnes.
  for x := 0 to LARGEUR-1 do
  begin
    result := result and VerifierZone(x, 0, x, HAUTEUR-1);
  end;

  // On vrifie les lignes.
  for y := 0 to HAUTEUR-1 do
  begin
    result := result and VerifierZone(0, y, LARGEUR-1, y);
  end;

  // On vrifie les blocs.
  for x := 0 to LGBLOC-1 do
  begin
    for y := 0 to HTBLOC-1 do
    begin
      result := result and VerifierZone(x*LGBLOC, y*HTBLOC, x*LGBLOC + LGBLOC-1, y*HTBLOC + HTBLOC-1);
    end;
  end;
end;

// Vrifie si la zone de la grille <x1, y1>  <x2, y2> respecte les rgles du sudoku :
// Elle ne doit pas contenir de valeurs plus d'une fois. Les cellules vides ne comptent pas.
// Cette mthode sert  la dfinition initiale de la grille, ainsi qu'au joueur qui tente de rsoudre
// la grille. Elle ne sert pas  la rsolution automatique de la grille. Elle n'a pas besoin d'tre
// optimise pour ses performances.
function TSudokuBase.VerifierZone(x1, y1, x2, y2: integer): boolean;
var
  x, y : integer;
  histo : array[VALEUR_MIN..VALEUR_MAX] of integer;
  valeur : integer;

begin
  fillchar(histo, sizeof(histo), 0);

  // Par dfaut, la zone est considre comme valide.
  result := true;
  for x := x1 to x2 do
  begin
    for y := y1 to y2 do
    begin
      valeur := ValeurCellule(x, y);
      if Valeur<>CELLULE_VIDE
      then begin
        inc(histo[Valeur]);
        if (histo[Valeur]>1)
        then begin
          // La valeur est prsente plus d'une fois. Ca suffit pour dire que la zone n'est pas valide.
          result := false;
          exit;
        end;
      end;
    end;
  end;
end;

initialization
  FListeAlgorithme := TStringList.Create;
  FListeAlgorithme.Sorted := true;
  FListeAlgorithme.CaseSensitive := false;
  FListeAlgorithme.Duplicates := dupIgnore;
finalization
  FListeAlgorithme.Free;
end.
