unit uSudokuNaif0;

interface
uses classes, sysutils, windows, uSudokuBase, uSudokuNaifBase;

type
  TSudokuNaif0 = class(TSudokuNaifBase)
  protected
    // Calcule les candidats possibles pour la cellule <x,y>
    procedure CandidatsPossibles(x, y : integer);

    // Recherche la cellule qui possde le moins de candidats possible et retourne ses coordonnes
    // dans <x, y>.
    // La procdure peut renvoyer :
    // <-1, -1> : Si toutes les cellules sont remplies.
    // <-2 -2> : S'il existe une cellule avec aucun candidat possible.
    procedure RechercherMinimumCandidat(var x, y : integer);
  public
    // Retourne le nom de l'algorithme utilis pour faire la rsolution.
    class function GetAlgorithmName : string; override;

    // Tente de rsoudre la grille depuis sa dfinition actuelle. En retour, la mthode renvoie :
    // - true : La grille a t rsolue avec succs !
    // - false : La grille ne possde pas de solution !
    function Resoudre(NbMaxSolution : integer) : boolean; override;

    // Initialise la grille avant de tanter sa rsolution.
    procedure InitialiserResolution; override;

    // Analyse la grille de jeu pour dterminer tous les coups possibles pour chaque cellule. Autrement
    // dit, pour chaque cellule, on regarde quelles sont les valeurs candidates.
    // ATTENTON : RechercherCoupsPossibles est la mthode la plus couteuse en temps d'excution.
    procedure RechercherTousLesCandidats; override;
  end;

implementation

// Calcule les candidats possibles pour la cellule <x,y>
procedure TSudokuNaif0.CandidatsPossibles(x, y : integer);
var
  Candidat : integer;
  i,j : integer;
  x1, y1 : integer;
begin
  // Par dfaut, tous les coups sont possibles.
  fillchar(FGrille[x, y].FlgCandidats, sizeof(FGrille[x, y].FlgCandidats), 1);

  // On essaie toutes les valeurs possibles et on regarde si chaque valeur est un candidat possible.
  for Candidat := VALEUR_MIN to VALEUR_MAX do
  begin
    // On regarde si la valeur n'est pas dj utilise sur le ligne.
    for i := 0 to LARGEUR-1 do
    begin
      if FGrille[i, y].Valeur = Candidat
      then FGrille[x, y].FlgCandidats[Candidat] := false; // Le candidat n'est pas possible !
    end;

    // On fait la mme chose en colonne.
    for i := 0 to LARGEUR-1 do
    begin
      if FGrille[x, i].Valeur = Candidat
      then FGrille[x, y].FlgCandidats[Candidat] := false; // Le candidat n'est pas possible !
    end;

    // Et on recommence pour la rgion
    x1 := (x div LGBLOC)*LGBLOC;
    y1 := (y div HTBLOC)*HTBLOC;
    for i := x1 to x1 +LGBLOC-1 do
    begin
      for j := y1 to y1+HTBLOC-1 do
      begin
        if FGrille[i, j].Valeur = Candidat
        then FGrille[x, y].FlgCandidats[Candidat] := false; // Le candidat n'est pas possible !
      end;
    end;
  end;
end;

class function TSudokuNaif0.GetAlgorithmName: string;
begin
  Result := 'Algorithme naf 0';
end;

// Initialise la grille avant de tanter sa rsolution.
procedure TSudokuNaif0.InitialiserResolution;
begin
  RechercherTousLesCandidats;
end;

// Recherche la cellule qui possde le moins de candidats possible et retourne ses coordonnes
// dans <x, y>.
// La procdure peut renvoyer :
// <-1, -1> : Si toutes les cellules sont remplies.
// <-2, -2> : S'il existe une cellule avec aucun candidat possible.
procedure TSudokuNaif0.RechercherMinimumCandidat(var x, y: integer);
var
  i, j : integer;
  TbPossible : array[VALEUR_MIN..VALEUR_MAX] of integer;
  CoupTrouve : array[VALEUR_MIN..VALEUR_MAX] of boolean;
  nb : integer;
  n : integer;
begin
  // Pour chaque cellule, on doit compter le nombre de candidats possibles, et renvoyer la premire
  // cellule qui possde le moins de candidats.
  Fillchar(CoupTrouve, sizeof(CoupTrouve), 0);

  for j := 0 to HAUTEUR-1 do
  begin
    for i := 0 to LARGEUR-1 do
    begin
      // Si la cellule n'est pas vide, on peut passer  la suivante immdiatement.
      if FGrille[i, j].Valeur<>CELLULE_VIDE
      then continue;

      // A prsent, on compte les candidats.
      nb := 0;
      for n := VALEUR_MIN to VALEUR_MAX do
      begin
        if (FGrille[i, j].FlgCandidats[n])
        then inc(nb);
      end;

      // Est-ce que la cellule possde au moin un candidat !
      if nb = 0
      then begin
        // La cellule n'a pas de solution. La grille est impossible !
        x := -2;
        y := -2;
        exit;
      end;

      if (nb > 0) and not CoupTrouve[nb]
      then begin
        // On vient de trouver la premire cellule avec Nb candidats possibles.
        tbPossible[nb] := (i shl 8) or j;
        CoupTrouve[nb] := true;
      end;
    end;
  end;

  // A prsent, on retourne la premire cellule ayant le plus petit nombre de candidats.
  for n := VALEUR_MIN to VALEUR_MAX do
  begin
    if CoupTrouve[n]
    then begin
      x := (tbPossible[n] and $ff00) shr 8;
      y := tbpossible[n] and $ff;
      exit;
    end;
  end;
  // Si on arrive ici, c'est qu'on n'a pas trouv de cellule vide avec au moin un candidat possible.
  // Comme le cas de la grille sans solution a dj t trait, c'est qu'il n'y a pas de cellule vide,
  // et donc que la grille est termine !
  x := -1;
  y := -1;
end;


procedure TSudokuNaif0.RechercherTousLesCandidats;
var
  x, y : integer;
begin
  for x := 0 to LARGEUR-1 do
  begin
    for y := 0 to HAUTEUR-1 do
    begin
      if FGrille[x, y].Valeur<>CELLULE_VIDE
      then continue;

      CandidatsPossibles(x, y); // On recherche les coups possible de la cellule.
    end;
  end;
end;

// Tente de rsoudre la grille depuis sa dfinition actuelle. En retour, la mthode renvoie :
// - true : La grille a t rsolue avec succs !
// - false : La grille ne possde pas de solution !
function TSudokuNaif0.Resoudre(NbMaxSolution : integer) : boolean;
var
  x, y : integer;
  n : integer;
  old : array[0..LARGEUR-1, 0..HAUTEUR-1] of TCellule;
begin
  result := true;

  // On commence par rechercher les candidats des cellules vides.
  RechercherTousLesCandidats;

  // Puis on recherche la premire cellule avec le moins de candidats
  RechercherMinimumCandidat(x, y);
  if (x=-1) and (y=-1)
  then begin
    result := true; // La rsolution est termine !
    MemoriserSolution; // On mmorise la solution trouve.
  end
  else begin
    if (x=-2) and (y=-2)
    then result := false // La grille n'a pas de solutions !
    else begin
      // On commence par sauvegarder la grille en cas de retour arrire
      move(FGrille, old, sizeof(FGrille));

      // Puis on examine les candidats de la cellule
      for n := VALEUR_MIN to VALEUR_MAX do
      begin
        if FGrille[x, y].FlgCandidats[n]
        then begin
          // On essaie la rsolution avec le candidat n
          FGrille[x, y].Valeur := n;

          // On continue recursivement.
          result := Resoudre(nbMaxSolution);

          // Si la grille a t rsolue et qu'on a atteind le nombre maximum de solutions autoris
          // on doit arrter la recherche. Ca vite de boucler trop longtemps pour les grilles avec un
          // trs grand nombre de solutions.
          if result and (ListeSolution.Count >= NbMaxSolution)
          then exit;

          // On vient de traiter le candidat n. On restaure la grille dans son tat initial pour
          // continuer avec le candidat suivant.
          move(old, FGrille, sizeof(FGrille));
        end;
      end;
    end;
  end;
end;

initialization
  TSudokuNaif0.RegisterAlgorithme;
end.
