unit uSudokuNaif3;

interface
uses classes, sysutils, windows, uSudokuBase, uSudokuNaifBase;

type
  TSudokuNaif3 = class(TSudokuNaifBase)
  private
    // Calcule les candidats possibles pour la cellule <x,y>
    procedure InterdireCandidats(x, y, valeur : integer);

    // Calcule les candidats possibles pour la cellule <x,y>
    procedure CandidatsPossibles(x, y : integer);

    // Recherche la cellule qui possde le moins de candidats possible et retourne ses coordonnes
    // dans <x, y>.
    // La procdure peut renvoyer :
    // <-1, -1> : Si toutes les cellules sont remplies.
    // <-2 -2> : S'il existe une cellule avec aucun candidat possible.
    procedure RechercherMinimumCandidat(var x, y : integer);
  public
    // Retourne le nom de l'algorithme utilis pour faire la rsolution.
    class function GetAlgorithmName : string; override;

    // Tente de rsoudre la grille depuis sa dfinition actuelle. En retour, la mthode renvoie :
    // - true : La grille a t rsolue avec succs !
    // - false : La grille ne possde pas de solution !
    function Resoudre(NbMaxSolution : integer) : boolean; override;

    // Initialise la grille avant de tanter sa rsolution.
    procedure InitialiserResolution; override;

    // Analyse la grille de jeu pour dterminer tous les coups possibles pour chaque cellule. Autrement
    // dit, pour chaque cellule, on regarde quelles sont les valeurs candidates.
    // ATTENTON : RechercherCoupsPossibles est la mthode la plus couteuse en temps d'excution.
    procedure RechercherTousLesCandidats; override;
  end;

implementation

procedure TSudokuNaif3.InterdireCandidats(x, y, valeur : integer);
var
  i,j : integer;
  x1, y1 : integer;
begin
  // On interdit la valeur fournie en paramtre
  for i := 0 to LARGEUR-1 do
  begin
    FGrille[i, y].FlgCandidats[valeur] := false;
    FGrille[x, i].FlgCandidats[Valeur] := false;
  end;

  // Et on recommence pour la rgion
  x1 := (x div LGBLOC)*LGBLOC;
  y1 := (y div HTBLOC)*HTBLOC;
  for i := x1 to x1 +LGBLOC-1 do
  begin
    for j := y1 to y1+HTBLOC-1 do
    begin
      FGrille[i,j].FlgCandidats[valeur] := false;
    end;
  end;
end;

// Cre et initialise une nouvelle instance de la classe TSudoku.
// La grille est initialise sur une grille vide.
procedure TSudokuNaif3.CandidatsPossibles(x, y: integer);
var
  i,j : integer;
  x1, y1 : integer;
  Cellule : PCellule;
begin
  // Par dfaut, tous les coups sont possibles.
  Cellule := @FGrille[x, y];

  fillchar(Cellule.FlgCandidats, sizeof(Cellule.FlgCandidats), 1);

  // On interdit les valeurs dj utilises en ligne ou en colonne.
  for i := 0 to LARGEUR-1 do
  begin
    Cellule.FlgCandidats[FGrille[i, y].Valeur] := false;
    Cellule.FlgCandidats[FGrille[x, i].Valeur] := false;
  end;

  // Et on recommence pour la rgion
  x1 := (x div LGBLOC)*LGBLOC;
  y1 := (y div HTBLOC)*HTBLOC;
  for i := x1 to x1 +LGBLOC-1 do
  begin
    for j := y1 to y1+HTBLOC-1 do
    begin
      Cellule.FlgCandidats[FGrille[i,j].Valeur] := false;
    end;
  end;
end;

class function TSudokuNaif3.GetAlgorithmName: string;
begin
  result := 'Algorithme naf 3';
end;

// Initialise la grille avant de tanter sa rsolution.
procedure TSudokuNaif3.InitialiserResolution;
begin
  RechercherTousLesCandidats;
end;

// Recherche la cellule qui possde le moins de candidats possible et retourne ses coordonnes
// dans <x, y>.
// La procdure peut renvoyer :
// <-1, -1> : Si toutes les cellules sont remplies.
// <-2, -2> : S'il existe une cellule avec aucun candidat possible.
procedure TSudokuNaif3.RechercherMinimumCandidat(var x, y: integer);
var
  i, j : integer;
  TbPossible : array[VALEUR_MIN..VALEUR_MAX] of integer;
  CoupTrouve : array[0..VALEUR_MAX] of boolean;
  nb : integer;
  n : integer;
  Cellule : PCellule;
begin
  // Pour chaque cellule, on doit compter le nombre de candidats possibles, et renvoyer la premire
  // cellule qui possde le moins de candidats.
  Fillchar(CoupTrouve, sizeof(CoupTrouve), 0);

  for j := 0 to HAUTEUR-1 do
  begin
    for i := 0 to LARGEUR-1 do
    begin
      Cellule := @FGrille[i, j];
      // Si la cellule n'est pas vide, on peut passer  la suivante immdiatement.
      if Cellule.Valeur<>CELLULE_VIDE
      then continue;

      // A prsent, on compte les candidats.
      nb := 0;
      for n := VALEUR_MIN to VALEUR_MAX do
      begin
        if (Cellule.FlgCandidats[n])
        then inc(nb);
      end;

      case nb of
      0: // La cellule n'a pas de solution. La grille est impossible !
        begin
          x := -2;
          y := -2;
          exit;
        end;
      1: // La cellule est vidente !
        begin
          // On joue la valeur
          for n := VALEUR_MIN to VALEUR_MAX do
          begin
            if (Cellule.FlgCandidats[n])
            then begin
              Cellule.Valeur := n;
              InterdireCandidats(i, j, n);
              break;
            end;
          end;
        end;
      else begin
        if not CoupTrouve[nb]
        then begin
          // On vient de trouver la premire cellule avec Nb candidats possibles.
          tbPossible[nb] := (i shl 8) or j;
          CoupTrouve[nb] := true;
        end;
      end;
      end;
    end;
  end;

  // A prsent, on retourne la premire cellule ayant le plus petit nombre de candidats.
  for n := VALEUR_MIN to VALEUR_MAX do
  begin
    if CoupTrouve[n]
    then begin
      x := (tbPossible[n] and $ff00) shr 8;
      y := tbpossible[n] and $ff;
      exit;
    end;
  end;
  // Si on arrive ici, c'est qu'on n'a pas trouv de cellule vide avec au moin un candidat possible.
  // Comme le cas de la grille sans solution a dj t trait, c'est qu'il n'y a pas de cellule vide,
  // et donc que la grille est termine !
  x := -1;
  y := -1;
end;

procedure TSudokuNaif3.RechercherTousLesCandidats;
var
  x, y : integer;
begin
  for x := 0 to LARGEUR-1 do
  begin
    for y := 0 to HAUTEUR-1 do
    begin
      if FGrille[x, y].Valeur<>CELLULE_VIDE
      then continue;

      CandidatsPossibles(x, y); // On recherche les coups possible de la cellule.
    end;
  end;
end;

// Tente de rsoudre la grille depuis sa dfinition actuelle. En retour, la mthode renvoie :
// - true : La grille a t rsolue avec succs !
// - false : La grille ne possde pas de solution !
function TSudokuNaif3.Resoudre(NbMaxSolution : integer) : boolean;
var
  x, y : integer;
  n : integer;
  old : array[0..LARGEUR-1, 0..HAUTEUR-1] of TCellule;
begin
  result := true;

  // On recherche la premire cellule avec le moins de candidats possibles.
  RechercherMinimumCandidat(x, y);
  if (x=-1) and (y=-1)
  then begin
    result := true; // La rsolution est termine !
    MemoriserSolution; // On mmorise la solution trouve.
  end
  else begin
    if (x=-2) and (y=-2)
    then result := false // La grille n'a pas de solutions !
    else begin
      // On commence par sauvegarder la grille pour pouvoir faire un retour arrire par la suite.
      move(FGrille, old, sizeof(FGrille));

      for n := VALEUR_MIN to VALEUR_MAX do
      begin
        if FGrille[x, y].FlgCandidats[n]
        then begin
          // On essaie la rsolution avec le candidat n
          FGrille[x, y].Valeur := n;
          InterdireCandidats(x, y, n); // On vient de placer une valeur, il faut appliquer
          // la modification des candidats possibles !

          // On continue recursivement.
          result := Resoudre(nbMaxSolution);

          // Si la grille a t rsolue et qu'on a atteind le nombre maximum de solutions autoris
          // on doit arrter la recherche. Ca vite de boucler trop longtemps pour les grilles avec un
          // trs grand nombre de solutions.
          if result and (ListeSolution.Count >= NbMaxSolution)
          then exit;

          // On vient de traiter le candidat n. On restaure la grille dans son tat initial pour
          // continuer avec le candidat suivant.
          move(old, FGrille, sizeof(FGrille));
        end;
      end;
    end;
  end;
end;

initialization
  TSudokuNaif3.RegisterAlgorithme;
end.
