unit uSudokuOptimise3;

interface
uses classes, sysutils, windows, uSudokuBase;

type
  TCellule = record
    Valeur : byte;     // Valeur de la cellule
    IsRevelee : boolean; // indique si la cellule est revlle : Sa valeur est fournie au dpart.
    FlgCandidats : word; // dsigne les marques dfinies par le joueur
  end;

  PCellule = ^TCellule;

  TSudokuOptimise3 = class(TSudokuBase)
  private
    FOwnSolution : boolean;
    // Dfinition de la grille de Sudoku
    FGrille : array[0..LARGEUR-1, 0..HAUTEUR-1] of TCellule;

    // Retourne le contenu de la grille sous la forme d'un chaine de LARGEUR*HAUTEUR caractres, avec
    // A = 0, B = 1 ...
    function DumpGrid : string;

    // Calcule les candidats possibles pour la cellule <x,y>
    procedure InterdireCandidats(x, y, valeur : integer); 

    // Calcule les candidats possibles pour la cellule <x,y>
    procedure CandidatsPossibles(x, y : integer);    

    // Ajoute la grille actuelle  la liste des solutions trouves.
    procedure MemoriserSolution;

    // Recherche la cellule qui possde le moins de candidats possible et retourne ses coordonnes
    // dans <x, y>.
    // La procdure peut renvoyer :
    // <-1, -1> : Si toutes les cellules sont remplies.
    // <-2 -2> : S'il existe une cellule avec aucun candidat possible.
    procedure RechercherMinimumCandidat(var x, y : integer);

  protected
    // Dfinit la valeur de la cellule <X,Y>  <Valeur>.
    // Cette mthode ne vrifie pas si le coup est valide.
    procedure DefinirValeur(X, Y, Valeur : integer); override;

    // Dfinit la cellule <X,Y> comme tant une celule rvle.
    procedure DefinirRevelee(X, Y : integer); override;
  public
    // Cre et initialise une nouvelle instance de la classe TGrilleSudoku.
    // La grille est initialise sur une grille vide.
    constructor Create; override;

    destructor Destroy; override;

    // Retourne le nom de l'algorithme utilis pour faire la rsolution.
    class function GetAlgorithmName : string; override;

    // Efface le contenu de la grille pour qu'on puisse lui dfinir une nouvelle
    // grille.
    procedure EffacerGrille; override;

    // Ajoute/supprime un marqueur de candidat possible pour la cellule <x, y>.
    // Cette mthode sert au joueur humain qui tente de rsoudre la grille.
    procedure MarkerCellule(x, y : integer; candidat : integer); override;

    // Indique si la cellule <X, Y> est une cellule rvle. Les cellules rvle ne peuvent pas tre
    // modifies et doivent tre dessine en noir.
    function IsRevelee(x, y : integer) : boolean; override;

    // Retourne la valeur de la cellule <X,Y>. Les cellules vides retournent CELLULE_VIDE.
    function ValeurCellule(X, Y : integer) : integer; override;

    // Indique si la valeur <Valeur> est dsigne comme tant un candidat pour la cellule <X,Y>.
    function IsCandidat(X, Y, Valeur : integer) : boolean; override;

    // Tente de rsoudre la grille depuis sa dfinition actuelle. En retour, la mthode renvoie :
    // - true : La grille a t rsolue avec succs !
    // - false : La grille ne possde pas de solution !
    function Resoudre(NbMaxSolution : integer) : boolean; override;

    // Initialise la grille avant de tanter sa rsolution.
    procedure InitialiserResolution; override;

    // Analyse la grille de jeu pour dterminer tous les coups possibles pour chaque cellule. Autrement
    // dit, pour chaque cellule, on regarde quelles sont les valeurs candidates.
    // ATTENTON : RechercherCoupsPossibles est la mthode la plus couteuse en temps d'excution.
    procedure RechercherTousLesCandidats; override;
  end;


implementation

const
  MasqueCandidats : array[0..16] of word =
    (word(not 0),
     word(not (1 shl  0)),
     word(not (1 shl  1)),
     word(not (1 shl  2)),
     word(not (1 shl  3)),
     word(not (1 shl  4)),
     word(not (1 shl  5)),
     word(not (1 shl  6)),
     word(not (1 shl  7)),
     word(not (1 shl  8)),
     word(not (1 shl  9)),
     word(not (1 shl 10)),
     word(not (1 shl 11)),
     word(not (1 shl 12)),
     word(not (1 shl 13)),
     word(not (1 shl 14)),
     word(not (1 shl 15)));

var
  NombreCandidats : array[0..$FFFF] of integer;
  ValeurEvidente : array[0..$FFFF] of cardinal;
  ParcoursCellNr : array[0..81] of integer;
  tbX : array[0..81] of integer;
  tbY : array[0..81] of integer;


procedure TSudokuOptimise3.InterdireCandidats(x, y, valeur : integer);
var
  i,j : integer;
  x1, y1 : integer;
  masque : word;
  CelluleLigne : pointer;
  CelluleColonne : pointer;
  CelluleRegion : pointer;
begin
  masque := MasqueCandidats[valeur];
  // On interdit la valeur fournie en paramtre
  CelluleLigne := @(FGrille[0, y].FlgCandidats);
  CelluleColonne := @(FGrille[x, 0].FlgCandidats);

  for i := 0 to LARGEUR-1 do
  begin
    word(CelluleLigne^) := word(CelluleLigne^) and masque;
    word(CelluleColonne^) := word(CelluleColonne^) and masque;
    inc(cardinal(CelluleLigne), HAUTEUR*sizeof(TCellule));
    inc(cardinal(CelluleColonne), sizeof(TCellule));
  end;

  // Et on recommence pour la rgion
  x1 := (x div LGBLOC)*LGBLOC;
  y1 := (y div HTBLOC)*HTBLOC;
  CelluleRegion := @(FGrille[x1,y1].FlgCandidats);
  for i := x1 to x1 +LGBLOC-1 do
  begin
    for j := y1 to y1+HTBLOC-1 do
    begin
      word(CelluleRegion^) := word(CelluleRegion^) and masque;
      inc(cardinal(CelluleRegion), sizeof(TCellule));
    end;
    inc(cardinal(CelluleRegion), sizeof(TCellule)*(HAUTEUR-HTBLOC));
  end;
end;

// Cre et initialise une nouvelle instance de la classe TSudoku.
// La grille est initialise sur une grille vide.
procedure TSudokuOptimise3.CandidatsPossibles(x, y: integer);
var
  i,j : integer;
  x1, y1 : integer;
  Cellule : PCellule;
begin
  // Par dfaut, tous les coups sont possibles.
  Cellule := @FGrille[x, y];

  Cellule.FlgCandidats := TOUS_LES_CANDIDATS;

  // On interdit les valeurs dj utilises en ligne ou en colonne.
  for i := 0 to LARGEUR-1 do
  begin
    Cellule.FlgCandidats := Cellule.FlgCandidats and
      MasqueCandidats[FGrille[i, y].Valeur] and
      MasqueCandidats[FGrille[x, i].Valeur];
  end;

  // Et on recommence pour la rgion
  x1 := (x div LGBLOC)*LGBLOC;
  y1 := (y div HTBLOC)*HTBLOC;
  for i := x1 to x1 +LGBLOC-1 do
  begin
    for j := y1 to y1+HTBLOC-1 do
    begin
      Cellule.FlgCandidats := Cellule.FlgCandidats and MasqueCandidats[FGrille[i,j].Valeur];
    end;
  end;
end;

constructor TSudokuOptimise3.Create;
begin
  inherited Create;
//  FillChar(FGrille, sizeof(FGrille), CELLULE_VIDE);
  FOwnSolution := true;
end;

// Dfinit la cellule <X,Y> comme tant une celule rvle.
procedure TSudokuOptimise3.DefinirRevelee(X, Y: integer);
begin
  FGrille[x, y].IsRevelee := true;
end;

// Dfinit la valeur de la cellule <X,Y>  <Valeur>.
// Cette mthode ne vrifie pas si le coup est valide.
procedure TSudokuOptimise3.DefinirValeur(X, Y, Valeur: integer);
begin
   FGrille[x, y].Valeur := valeur;
end;

destructor TSudokuOptimise3.Destroy;
begin
  if not FOwnSolution
  then ListeSolution := nil;

  inherited Destroy;
end;

// Retourne le contenu de la grille sous la forme d'un chaine de LARGEUR*HAUTEUR caractres, avec
// A = 0, B = 1 ...
function TSudokuOptimise3.DumpGrid: string;
var
  i,j : integer;
  n : integer;
begin
  n := 1;
  SetLength(result, LARGEUR*HAUTEUR);
  for j := 0 to HAUTEUR-1 do
  begin
    for i := 0 to LARGEUR-1 do
    begin
      result[n] := chr(FGrille[i,j].Valeur + 65);
      inc(n);
    end;
  end;
end;

procedure TSudokuOptimise3.EffacerGrille;
begin
  FillChar(FGrille, sizeof(FGrille), 0);
end;

class function TSudokuOptimise3.GetAlgorithmName: string;
begin
  result := 'Algorithme Optimis 3';
end;

// Initialise la grille avant de tanter sa rsolution.
procedure TSudokuOptimise3.InitialiserResolution;
begin
  RechercherTousLesCandidats;
end;

// Indique si la valeur <Valeur> est dsigne comme tant un candidat pour la cellule <X,Y>.
function TSudokuOptimise3.IsCandidat(X, Y, Valeur: integer): boolean;
begin
  result := (FGrille[x, y].FlgCandidats and (1 shl (valeur-1)))<>0;
end;

// Indique si la cellule <X, Y> est une cellule rvle. Les cellules rvle ne peuvent pas tre
// modifies et doivent tre dessine en noir.
function TSudokuOptimise3.IsRevelee(x, y: integer): boolean;
begin
  result := FGrille[x, y].IsRevelee;
end;

// Ajoute/supprime un marqueur de valeur possible pour la cellule <x, y>.
// Cette mthode sert au joueur humain qui tente de rsoudre la grille.
procedure TSudokuOptimise3.MarkerCellule(x, y, candidat: integer);
begin
  if not FGrille[x, y].IsRevelee
  then FGrille[x,y].FlgCandidats := FGrille[x,y].FlgCandidats xor (1 shl (candidat-1));
end;

// Ajoute la grille actuelle  la liste des solutions trouves.
procedure TSudokuOptimise3.MemoriserSolution;
var
  t0 : int64;
  s : string;
begin
  // Tout d'abord, on mmorise la date  laquelle la solution a t trouve.
  QueryPerformanceCounter(t0);

  s := DumpGrid;
  s := s + ' - ' + IntToStr(t0); // On ajoute la date de rsolution  la grille trouve.
  ListeSolution.Add(s);
end;

procedure TSudokuOptimise3.RechercherMinimumCandidat(var x, y: integer);
type
  TbCellule = array[0..LARGEUR*HAUTEUR-1] of TCellule;
var
  i : integer;
  TbPossible : array[VALEUR_MIN..VALEUR_MAX] of integer;
  CoupTrouve : array[0..VALEUR_MAX] of boolean;
  nb : integer;
  n : integer;
  Cellule : PCellule;
  ptrCellNo : ^integer;
begin
  // Pour chaque cellule, on doit compter le nombre de candidats possibles, et renvoyer la premire
  // cellule qui possde le moins de candidats.
  Fillchar(CoupTrouve, sizeof(CoupTrouve), 0);

  ptrCellNo := @ParcoursCellNr[0];
  for i := 0 to LARGEUR*HAUTEUR-1 do
  begin
    Cellule := @tbCellule(FGrille)[ptrCellNo^];
    // Si la cellule n'est pas vide, on peut passer  la suivante immdiatement.
    if Cellule.Valeur=CELLULE_VIDE
    then begin
      // A prsent, on compte les candidats.
      nb := NombreCandidats[Cellule.FlgCandidats];

      case nb of
      0: // La cellule n'a pas de solution. La grille est impossible !
        begin
          x := -2;
          y := -2;
          exit;
        end;
      1: // La cellule est vidente !
        begin
          // On joue la valeur
          n :=  ValeurEvidente[Cellule.FlgCandidats];
          Cellule.Valeur := n;
          InterdireCandidats(ptrCellNo^ div HAUTEUR, ptrCellNo^ mod HAUTEUR, n);
        end;
      else begin
        if not CoupTrouve[nb]
        then begin
          // On vient de trouver la premire cellule avec Nb candidats possibles.
          tbPossible[nb] := ptrCellNo^;
          CoupTrouve[nb] := true;
        end;
      end;
      end;
    end;
    inc(cardinal(ptrCellNo), 4);
  end;

  // A prsent, on retourne la premire cellule ayant le plus petit nombre de candidats.
  for n := VALEUR_MIN to VALEUR_MAX do
  begin
    if CoupTrouve[n]
    then begin
      x := tbPossible[n] div HAUTEUR;
      y := tbpossible[n] mod HAUTEUR;
      exit;
    end;
  end;
  // Si on arrive ici, c'est qu'on n'a pas trouv de cellule vide avec au moin un candidat possible.
  // Comme le cas de la grille sans solution a dj t trait, c'est qu'il n'y a pas de cellule vide,
  // et donc que la grille est termine !
  x := -1;
  y := -1;
end;


procedure TSudokuOptimise3.RechercherTousLesCandidats;
var
  x, y : integer;
begin
  for x := 0 to LARGEUR-1 do
  begin
    for y := 0 to HAUTEUR-1 do
    begin
      if FGrille[x, y].Valeur<>CELLULE_VIDE
      then continue;

      CandidatsPossibles(x, y); // On recherche les coups possible de la cellule.
    end;
  end;
end;

// Tente de rsoudre la grille depuis sa dfinition actuelle. En retour, la mthode renvoie :
// - true : La grille a t rsolue avec succs !
// - false : La grille ne possde pas de solution !
function TSudokuOptimise3.Resoudre(NbMaxSolution : integer) : boolean;
var
  x, y : integer;
  n : integer;
  old : array[0..LARGEUR-1, 0..HAUTEUR-1] of TCellule;
  masque : cardinal;
begin
  result := true;

  // On recherche la premire cellule avec le moins de candidats possibles.
  RechercherMinimumCandidat(x, y);

  if (x=-1) and (y=-1)
  then begin
    result := true; // La rsolution est termine !
    MemoriserSolution; // On mmorise la solution trouve.
  end
  else begin
    if (x=-2) and (y=-2)
    then result := false // La grille n'a pas de solutions !
    else begin
      // On commence par sauvegarder la grille pour pouvoir faire un retour arrire par la suite.
      move(FGrille, old, sizeof(FGrille));

      masque := 1;
      for n := VALEUR_MIN to VALEUR_MAX do
      begin
        if (FGrille[x, y].FlgCandidats and masque)<>0
        then begin
          // On essaie la rsolution avec le candidat n
          FGrille[x, y].Valeur := n;
          InterdireCandidats(x, y, n); // On vient de placer une valeur, il faut appliquer
          // la modification des candidats possibles !

          // On continue recursivement.
          result := Resoudre(nbMaxSolution);

          // Si la grille a t rsolue et qu'on a atteind le nombre maximum de solutions autoris
          // on doit arrter la recherche. Ca vite de boucler trop longtemps pour les grilles avec un
          // trs grand nombre de solutions.
          if result and (ListeSolution.Count >= NbMaxSolution)
          then exit;

          // On vient de traiter le candidat n. On restaure la grille dans son tat initial pour
          // continuer avec le candidat suivant.
          move(old, FGrille, sizeof(FGrille));
        end;
        masque := masque shl 1;
      end;
    end;
  end;
end;

// Retourne la valeur de la cellule <X,Y>. Les cellules vides retournent CELLULE_VIDE.
function TSudokuOptimise3.ValeurCellule(X, Y: integer): integer;
begin
  result := FGrille[x, y].Valeur;
end;

procedure PreCalculerNombreCandidats;
var
  i, j: integer;
  nb : integer;
  n : cardinal;
begin
  // Initialisation du tableau NbMarques. Ce tableau a pour but de nous viter de compter le nombre de
  // candidats pour chaque cellules : Les candidats sont indiqus sous la forme d'un masque de bits
  // dans un WORD (16 bits).
  // On calcule les 65536 entres possibles, et pour chacune d'entre elle, on compte le nombre de bits
  //  1 correspondant.
  for i := 0 to high(NombreCandidats) do
  begin
    nb := 0;  // Pour l'instant, il n'y a aucun bit positinn sur la valeur.
    j := i;
    while (j<>0) do
    begin
      if (j and 1)<>0 // On regarde si le bit de poid faible est positionn.
      then inc(nb);
      j := j shr 1;
    end;
    NombreCandidats[i] := nb; // On mmorise le rsultat trouv.
    ValeurEvidente[i] := 0;
    if nb = 1  // On est dans le cas d'une cellule vidente. On fait prcalcule pour savoir
               // de laquelle il s'agit.
    then begin
      nb := i;
      while nb<>0 do
      begin
        inc(ValeurEvidente[i]);
        nb := nb shr 1;
      end;
    end;
  end;

  n := 0;
  for j := 0 to HAUTEUR-1 do
  begin
    for i := 0 to LARGEUR-1 do
    begin
      ParcoursCellNr[n] := i*HAUTEUR + j;
      tbX[i*HAUTEUR + j] := i;
      tbY[i*HAUTEUR + j] := j;
      inc(n);
    end;
  end;
end;

initialization
  PreCalculerNombreCandidats;
  TSudokuOptimise3.RegisterAlgorithme;
end.
