unit uSudokuOptimiseFinal;

interface
uses classes, sysutils, windows, uSudokuBase;

type
  PShortint = ^Shortint;
  TCellule = record
    Valeur : byte;     // Valeur de la cellule
    cellNo : byte; // indique si la cellule est revlle : Sa valeur est fournie au dpart.
    FlgCandidats : word; // dsigne les marques dfinies par le joueur
  end;
  TbCellule = array[0..LARGEUR*HAUTEUR-1] of TCellule;

  PCellule = ^TCellule;

  TSudokuOptimiseFinal = class(TSudokuBase)
  private
    // Dfinition de la grille de Sudoku
    FGrille : array[0..LARGEUR-1, 0..HAUTEUR-1] of TCellule;
    FVide : array[0..LARGEUR-1, 0..HAUTEUR-1] of boolean;
    FRevele : array[0..LARGEUR-1, 0..HAUTEUR-1] of boolean;

    FOwnSolution : boolean;
    // Retourne le contenu de la grille sous la forme d'un chaine de LARGEUR*HAUTEUR caractres, avec
    // A = 0, B = 1 ...
    function DumpGrid : string;

    // Calcule les candidats possibles pour la cellule <x,y>
    function InterdireCandidats(x, y, valeur : integer) : boolean;

    // Calcule les candidats possibles pour la cellule <x,y>
    procedure CandidatsPossibles(x, y : integer);

    // Ajoute la grille actuelle  la liste des solutions trouves.
    procedure MemoriserSolution;

    // Recherche la cellule qui possde le moins de candidats possible et retourne ses coordonnes
    // dans <x, y>.
    // La procdure peut renvoyer :
    // <-1, -1> : Si toutes les cellules sont remplies.
    // <-2 -2> : S'il existe une cellule avec aucun candidat possible.
    function RechercherMinimumCandidat : integer;

  protected
    // Dfinit la valeur de la cellule <X,Y>  <Valeur>.
    // Cette mthode ne vrifie pas si le coup est valide.
    procedure DefinirValeur(X, Y, Valeur : integer); override;

    // Dfinit la cellule <X,Y> comme tant une celule rvle.
    procedure DefinirRevelee(X, Y : integer); override;
  public
    // Cre et initialise une nouvelle instance de la classe TGrilleSudoku.
    // La grille est initialise sur une grille vide.
    constructor Create; override;

    destructor Destroy; override;

    // Retourne le nom de l'algorithme utilis pour faire la rsolution.
    class function GetAlgorithmName : string; override;

    // Efface le contenu de la grille pour qu'on puisse lui dfinir une nouvelle
    // grille.
    procedure EffacerGrille; override;
    
    // Ajoute/supprime un marqueur de candidat possible pour la cellule <x, y>.
    // Cette mthode sert au joueur humain qui tente de rsoudre la grille.
    procedure MarkerCellule(x, y : integer; candidat : integer); override;

    // Indique si la cellule <X, Y> est une cellule rvle. Les cellules rvle ne peuvent pas tre
    // modifies et doivent tre dessine en noir.
    function IsRevelee(x, y : integer) : boolean; override;

    // Retourne la valeur de la cellule <X,Y>. Les cellules vides retournent CELLULE_VIDE.
    function ValeurCellule(X, Y : integer) : integer; override;

    // Indique si la valeur <Valeur> est dsigne comme tant un candidat pour la cellule <X,Y>.
    function IsCandidat(X, Y, Valeur : integer) : boolean; override;

    // Tente de rsoudre la grille depuis sa dfinition actuelle. En retour, la mthode renvoie :
    // - true : La grille a t rsolue avec succs !
    // - false : La grille ne possde pas de solution !
    function Resoudre(NbMaxSolution : integer) : boolean; override;

    // Initialise la grille avant de tanter sa rsolution.
    procedure InitialiserResolution; override;

    // Analyse la grille de jeu pour dterminer tous les coups possibles pour chaque cellule. Autrement
    // dit, pour chaque cellule, on regarde quelles sont les valeurs candidates.
    // ATTENTON : RechercherCoupsPossibles est la mthode la plus couteuse en temps d'excution.
    procedure RechercherTousLesCandidats; override;
  end;

const
  MasqueCandidats : array[0..16] of word =
    (word(not 0),
     word(not (1 shl  0)),
     word(not (1 shl  1)),
     word(not (1 shl  2)),
     word(not (1 shl  3)),
     word(not (1 shl  4)),
     word(not (1 shl  5)),
     word(not (1 shl  6)),
     word(not (1 shl  7)),
     word(not (1 shl  8)),
     word(not (1 shl  9)),
     word(not (1 shl 10)),
     word(not (1 shl 11)),
     word(not (1 shl 12)),
     word(not (1 shl 13)),
     word(not (1 shl 14)),
     word(not (1 shl 15)));

const
  RegionX: array[0..LARGEUR-1] of integer =
    ((0 div LGBLOC)*LGBLOC,
     (1 div LGBLOC)*LGBLOC,
     (2 div LGBLOC)*LGBLOC,
     (3 div LGBLOC)*LGBLOC,
     (4 div LGBLOC)*LGBLOC,
     (5 div LGBLOC)*LGBLOC,
     (6 div LGBLOC)*LGBLOC,
     (7 div LGBLOC)*LGBLOC,
     (8 div LGBLOC)*LGBLOC);
  RegionY: array[0..HAUTEUR-1] of integer =
    ((0 div HTBLOC)*HTBLOC,
     (1 div HTBLOC)*HTBLOC,
     (2 div HTBLOC)*HTBLOC,
     (3 div HTBLOC)*HTBLOC,
     (4 div HTBLOC)*HTBLOC,
     (5 div HTBLOC)*HTBLOC,
     (6 div HTBLOC)*HTBLOC,
     (7 div HTBLOC)*HTBLOC,
     (8 div HTBLOC)*HTBLOC);

implementation

var
  NombreCandidats : array[0..$FFFF] of integer;
  ValeurEvidente : array[0..$FFFF] of cardinal;
  ParcoursCellNr : array[0..81] of integer=
  (  19,
     22,
     1,
     28,
     46,
     55,
     64,
     72,
     47,
     38,
     36,
     65,
     74,
     77,
     27,
     54,
     18,
     58,
     23,
     59,
     68,
     24,
     9,
     25,
     2,
     11,
     31,
     49,
     67,
     75,
     4,
     66,
     39,
     48,
     3,
     12,
     14,
     32,
     41,
     42,
     44,
     71,
     43,
     70,
     17,
     80,
     15,
     51,
     60,
     78,
     52,
     61,
     35,
     53,
     33,
     7,
     34,
     8,
     0,
     45,
     63,
     10,
     37,
     73,
     20,
     29,
     56,
     21,
     30,
     57,
     13,
     40,
     76,
     5,
     50,
     6,
     69,
     16,
     79,
     26,
     62,
     -1
  );

  tbX : array[0..81] of integer;
  tbY : array[0..81] of integer;
  nbResoudre : int64;

function TSudokuOptimiseFinal.InterdireCandidats(x, y, valeur : integer) : boolean;
var
  x1, y1 : integer;
  masque : word;
  CelluleLigne : pointer;
  CelluleColonne : pointer;
  CelluleRegion : pointer;

label Fin;
begin
  masque := MasqueCandidats[valeur];
  // On interdit la valeur fournie en paramtre
  CelluleLigne := @(FGrille[0, y]);
  CelluleColonne := @(FGrille[x, 0]);

  tbCellule(CelluleLigne^)[0].FlgCandidats := tbCellule(CelluleLigne^)[0].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleColonne^)[0].FlgCandidats :=
    tbCellule(CelluleColonne^)[0].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleLigne^)[HAUTEUR].FlgCandidats :=
    tbCellule(CelluleLigne^)[HAUTEUR].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleColonne^)[1].FlgCandidats :=
    tbCellule(CelluleColonne^)[1].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleLigne^)[HAUTEUR*2].FlgCandidats :=
    tbCellule(CelluleLigne^)[HAUTEUR*2].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleColonne^)[2].FlgCandidats :=
    tbCellule(CelluleColonne^)[2].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleLigne^)[HAUTEUR*3].FlgCandidats :=
    tbCellule(CelluleLigne^)[HAUTEUR*3].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleColonne^)[3].FlgCandidats :=
    tbCellule(CelluleColonne^)[3].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleLigne^)[HAUTEUR*4].FlgCandidats :=
    tbCellule(CelluleLigne^)[HAUTEUR*4].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleColonne^)[4].FlgCandidats :=
    tbCellule(CelluleColonne^)[4].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleLigne^)[HAUTEUR*5].FlgCandidats :=
    tbCellule(CelluleLigne^)[HAUTEUR*5].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleColonne^)[5].FlgCandidats :=
    tbCellule(CelluleColonne^)[5].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleLigne^)[HAUTEUR*6].FlgCandidats :=
    tbCellule(CelluleLigne^)[HAUTEUR*6].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleColonne^)[6].FlgCandidats :=
    tbCellule(CelluleColonne^)[6].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleLigne^)[HAUTEUR*7].FlgCandidats :=
    tbCellule(CelluleLigne^)[HAUTEUR*7].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleColonne^)[7].FlgCandidats :=
    tbCellule(CelluleColonne^)[7].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleLigne^)[HAUTEUR*8].FlgCandidats :=
    tbCellule(CelluleLigne^)[HAUTEUR*8].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleColonne^)[8].FlgCandidats :=
    tbCellule(CelluleColonne^)[8].FlgCandidats and masque;
  asm jz Fin end;

  x1 := RegionX[x];
  y1 := RegionY[y];

  CelluleRegion := @(FGrille[x1,y1]);
  tbCellule(CelluleRegion^)[0].FlgCandidats :=
    tbCellule(CelluleRegion^)[0].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleRegion^)[1].FlgCandidats :=
    tbCellule(CelluleRegion^)[1].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleRegion^)[2].FlgCandidats :=
    tbCellule(CelluleRegion^)[2].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleRegion^)[HAUTEUR + 0].FlgCandidats :=
    tbCellule(CelluleRegion^)[HAUTEUR + 0].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleRegion^)[HAUTEUR + 1].FlgCandidats :=
    tbCellule(CelluleRegion^)[HAUTEUR + 1].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleRegion^)[HAUTEUR + 2].FlgCandidats :=
    tbCellule(CelluleRegion^)[HAUTEUR + 2].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleRegion^)[HAUTEUR*2 + 0].FlgCandidats :=
    tbCellule(CelluleRegion^)[HAUTEUR*2 + 0].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleRegion^)[HAUTEUR*2 + 1].FlgCandidats :=
    tbCellule(CelluleRegion^)[HAUTEUR*2 + 1].FlgCandidats and masque;
  asm jz Fin end;

  tbCellule(CelluleRegion^)[HAUTEUR*2 + 2].FlgCandidats :=
    tbCellule(CelluleRegion^)[HAUTEUR*2 + 2].FlgCandidats and masque;
  asm jz Fin end;
  
  result := true;
  exit;
fin:
  result := false;
end;

// Cre et initialise une nouvelle instance de la classe TSudoku.
// La grille est initialise sur une grille vide.
procedure TSudokuOptimiseFinal.CandidatsPossibles(x, y: integer);
var
  i,j : integer;
  x1, y1 : integer;
  Cellule : PCellule;
begin
  // Par dfaut, tous les coups sont possibles.
  Cellule := @FGrille[x, y];

  Cellule.FlgCandidats := TOUS_LES_CANDIDATS;

  // On interdit les valeurs dj utilises en ligne ou en colonne.
  for i := 0 to LARGEUR-1 do
  begin
    Cellule.FlgCandidats := Cellule.FlgCandidats and
      MasqueCandidats[FGrille[i, y].Valeur] and
      MasqueCandidats[FGrille[x, i].Valeur];
  end;

  // Et on recommence pour la rgion
  x1 := (x div LGBLOC)*LGBLOC;
  y1 := (y div HTBLOC)*HTBLOC;
  for i := x1 to x1 +LGBLOC-1 do
  begin
    for j := y1 to y1+HTBLOC-1 do
    begin
      Cellule.FlgCandidats := Cellule.FlgCandidats and MasqueCandidats[FGrille[i,j].Valeur];
    end;
  end;
end;

constructor TSudokuOptimiseFinal.Create;
begin
  inherited Create;
//  FillChar(FGrille, sizeof(FGrille), CELLULE_VIDE);
  FOwnSolution := true;
end;

// Dfinit la cellule <X,Y> comme tant une celule rvle.
procedure TSudokuOptimiseFinal.DefinirRevelee(X, Y: integer);
begin
  FRevele[x, y] := true;
end;

// Dfinit la valeur de la cellule <X,Y>  <Valeur>.
// Cette mthode ne vrifie pas si le coup est valide.
procedure TSudokuOptimiseFinal.DefinirValeur(X, Y, Valeur: integer);
begin
   FGrille[x, y].Valeur := valeur;
end;

destructor TSudokuOptimiseFinal.Destroy;
begin
  if not FOwnSolution
  then ListeSolution := nil;

  inherited Destroy;
end;

// Retourne le contenu de la grille sous la forme d'un chaine de LARGEUR*HAUTEUR caractres, avec
// A = 0, B = 1 ...
function TSudokuOptimiseFinal.DumpGrid: string;
var
  i,j : integer;
  n : integer;
begin
  n := 1;
  SetLength(result, LARGEUR*HAUTEUR);
  for j := 0 to HAUTEUR-1 do
  begin
    for i := 0 to LARGEUR-1 do
    begin
      result[n] := chr(FGrille[i,j].Valeur + 65);
      inc(n);
    end;
  end;
end;

procedure TSudokuOptimiseFinal.EffacerGrille;
begin
  FillChar(FGrille, sizeof(FGrille), 0);
  FillChar(FVide, sizeof(FVide), 0);
  FillChar(FRevele, sizeof(FRevele), 0);
end;

class function TSudokuOptimiseFinal.GetAlgorithmName: string;
begin
  result := 'Algorithme optimis Final';
end;

// Initialise la grille avant de tanter sa rsolution.
procedure TSudokuOptimiseFinal.InitialiserResolution;
begin
  RechercherTousLesCandidats;
  nbResoudre := 0;
end;

// Indique si la valeur <Valeur> est dsigne comme tant un candidat pour la cellule <X,Y>.
function TSudokuOptimiseFinal.IsCandidat(X, Y, Valeur: integer): boolean;
begin
  result := (FGrille[x, y].FlgCandidats and (1 shl (valeur-1)))<>0;
end;

// Indique si la cellule <X, Y> est une cellule rvle. Les cellules rvle ne peuvent pas tre
// modifies et doivent tre dessine en noir.
function TSudokuOptimiseFinal.IsRevelee(x, y: integer): boolean;
begin
  result := FRevele[x, y];
end;

// Ajoute/supprime un marqueur de valeur possible pour la cellule <x, y>.
// Cette mthode sert au joueur humain qui tente de rsoudre la grille.
procedure TSudokuOptimiseFinal.MarkerCellule(x, y, candidat: integer);
begin
  if not FRevele[x, y]
  then FGrille[x,y].FlgCandidats := FGrille[x,y].FlgCandidats xor (1 shl (candidat-1));
end;

// Ajoute la grille actuelle  la liste des solutions trouves.
procedure TSudokuOptimiseFinal.MemoriserSolution;
var
  t0 : int64;
  s : string;
begin
  // Tout d'abord, on mmorise la date  laquelle la solution a t trouve.
  QueryPerformanceCounter(t0);

  s := DumpGrid;
  s := s + ' - ' + IntToStr(t0); // On ajoute la date de rsolution  la grille trouve.
  ListeSolution.Add(s);
end;

function TSudokuOptimiseFinal.RechercherMinimumCandidat : integer;
var
  TbPossible : array[VALEUR_MIN..VALEUR_MAX] of integer;
  CoupTrouve : array[0..VALEUR_MAX] of boolean;
  nb : integer;
  n : integer;
  Cellule : PCellule;
  ptrCellNo : PShortint;
  ptrCellNo2 : PShortint;
  CellNo : byte;
begin
  // Pour chaque cellule, on doit compter le nombre de candidats possibles, et renvoyer la premire
  // cellule qui possde le moins de candidats.
  Fillchar(CoupTrouve, sizeof(CoupTrouve), 0);

  ptrCellNo := @(FGrille[0,0].CellNo);
  ptrCellNo2 := ptrCellNo;
  while (ptrCellNo^>=0) do
  begin
    CellNo := ptrCellNo^;
    Cellule := @tbCellule(FGrille)[CellNo];
    // Si la cellule n'est pas vide, on peut passer  la suivante immdiatement.
    if Cellule.Valeur=CELLULE_VIDE
    then begin
      ptrCellNo2^ := CellNo;
      inc(ptrCellNo2, 4);
      // A prsent, on compte les candidats.
      nb := NombreCandidats[Cellule.FlgCandidats];
      if nb = 1
      then begin
        // On joue la valeur
        n :=  ValeurEvidente[Cellule.FlgCandidats];
        Cellule.Valeur := n;
        Cellule.FlgCandidats := $FFFF;
        if not InterdireCandidats(tbX[CellNo], tbY[CellNo], n)
        then begin
          result := -2;
          exit;
        end;
      end
      else begin
        if not CoupTrouve[nb]
        then begin
          // On vient de trouver la premire cellule avec Nb candidats possibles.
          tbPossible[nb] := CellNo;
          CoupTrouve[nb] := true;
        end;
      end;
    end;
    inc(cardinal(ptrCellNo), 4);
  end;
  ptrCellNo2^ := -1;
  // A prsent, on retourne la premire cellule ayant le plus petit nombre de candidats.
  for n := VALEUR_MIN to VALEUR_MAX do
  begin
    if CoupTrouve[n]
    then begin
      result := tbPossible[n];
      exit;
    end;
  end;
  // Si on arrive ici, c'est qu'on n'a pas trouv de cellule vide avec au moin un candidat possible.
  // Comme le cas de la grille sans solution a dj t trait, c'est qu'il n'y a pas de cellule vide,
  // et donc que la grille est termine !
  result := -1;
end;

procedure TSudokuOptimiseFinal.RechercherTousLesCandidats;
var
  x, y : integer;
  Cellule : PCellule;
  ptrCellNo : PShortint;
  n : integer;
begin
  ptrCellNo := @(FGrille[0,0].CellNo);
  n := 0;
  for y := 0 to HAUTEUR-1 do
  begin
    for x := 0 to LARGEUR-1 do
    begin
      ptrCellNo^ := ParcoursCellNr[n];
      inc(cardinal(ptrCellNo), 4);      
      inc(n);
      Cellule := @FGrille[x, y];
      if Cellule.Valeur<>CELLULE_VIDE
      then begin
        Cellule.FlgCandidats := $FFFF;
      end
      else begin //continue;
        CandidatsPossibles(x, y); // On recherche les coups possible de la cellule.
      end;
    end;
  end;
  ptrCellNo^ := -1;
end;


function TSudokuOptimiseFinal.Resoudre(NbMaxSolution : integer) : boolean;
var
  x, y : integer;
  n : integer;
  old : array[0..LARGEUR-1, 0..HAUTEUR-1] of TCellule;
  masque : cardinal;
  CelluleNo : integer;
  Cellule : PCellule;
begin
  inc(nbResoudre);
  // On recherche la premire cellule avec le moins de candidats possibles.
  CelluleNo := RechercherMinimumCandidat;

  case CelluleNo of
  -1: begin
        MemoriserSolution; // On mmorise la solution trouve.
        result := ListeSolution.Count >= NbMaxSolution;
      end;
  -2: // La grille n'a pas de solutions !
      result := false;
  else begin
    // On commence par sauvegarder la grille pour pouvoir faire un retour arrire par la suite.
    move(FGrille, old, sizeof(FGrille));
    x := tbX[CelluleNo];
    y := tbY[CelluleNo];
    Cellule := @(tbCellule(FGrille)[CelluleNo]);
    result := false;

    masque := Cellule.FlgCandidats;
    n := 1;
    repeat
      if masque and 1<>0
      then begin
        Cellule.Valeur := n;
        Cellule.FlgCandidats := $FFFF;
        if InterdireCandidats(x, y, n)
        then begin
          result := Resoudre(nbMaxSolution);
          if result
          then exit;
        end
        else result := false;

        // On vient de traiter le candidat n. On restaure la grille dans son tat initial pour
        // continuer avec le candidat suivant.
        move(old, FGrille, sizeof(FGrille));
      end;
      inc(n);
      masque := masque shr 1;
    until masque = 0;
  end;
  end;
end;            

// Retourne la valeur de la cellule <X,Y>. Les cellules vides retournent CELLULE_VIDE.
function TSudokuOptimiseFinal.ValeurCellule(X, Y: integer): integer;
begin
  result := FGrille[x, y].Valeur;
end;

procedure PreCalculerNombreCandidats;
var
  i, j: integer;
  nb : integer;
begin
  // Initialisation du tableau NbMarques. Ce tableau a pour but de nous viter de compter le nombre de
  // candidats pour chaque cellules : Les candidats sont indiqus sous la forme d'un masque de bits
  // dans un WORD (16 bits).
  // On calcule les 65536 entres possibles, et pour chacune d'entre elle, on compte le nombre de bits
  //  1 correspondant.
  for i := 0 to high(NombreCandidats) do
  begin
    nb := 0;  // Pour l'instant, il n'y a aucun bit positinn sur la valeur.
    j := i;
    while (j<>0) do
    begin
      if (j and 1)<>0 // On regarde si le bit de poid faible est positionn.
      then inc(nb);
      j := j shr 1;
    end;
    NombreCandidats[i] := nb; // On mmorise le rsultat trouv.
    ValeurEvidente[i] := 0;
    if nb = 1  // On est dans le cas d'une cellule vidente. On fait prcalcule pour savoir
               // de laquelle il s'agit.
    then begin
      nb := i;
      while nb<>0 do
      begin
        inc(ValeurEvidente[i]);
        nb := nb shr 1;
      end;
    end;
  end;

  for j := 0 to HAUTEUR-1 do
  begin
    for i := 0 to LARGEUR-1 do
    begin
      tbX[i*HAUTEUR + j] := i;
      tbY[i*HAUTEUR + j] := j;
    end;
  end;
end;

initialization
  PreCalculerNombreCandidats;
  TSudokuOptimiseFinal.RegisterAlgorithme;
end.
